# -*- coding: utf-8 -*-
##############################################################################
# This file is part of the "use_control" OpenERP module.
# See LICENCE file at the root of the module folder.
##############################################################################

import gc
import pooler
import time
import netsvc
import os
import os.path
import tools
from math import ceil
from psycopg2 import IntegrityError
from service import security
import logging

logger = logging.getLogger('use_control')

##############################################################################
# Service to record use of the database
##############################################################################

objects_proxy = type(netsvc.ExportService._services['object'])

class recording_objects_proxy(objects_proxy):
    access_cache = {}
    HOUR_MINI = 1.0
    UNTIMED_CALLS = {
        # values are either '*' or a list of methods
        ## 'foo': ['bar', 'baz'],
        'res.request': '*',
    }

    # f*cking API ಠ_ಠ
    def dispatch(self, func, auth, params):
        if func == 'execute':
            db, uid, _passwd, model, method = params[:5]
            self._log_access(db, uid, model, method)
        return super(recording_objects_proxy, self).dispatch(func, auth, params)

    def _log_access(self, db, uid, model, method):
        cr = pooler.get_db_only(db).cursor()
        try:
            # check if database is blocked
            cr.execute("select count(1) from information_schema.tables where table_name=%s", ('use_control_db_block',))
            if not cr.fetchone()[0]:
                logger.info('[%s] No table named "use_control_db_block"', db)
                return
            cr.execute('SELECT name FROM use_control_db_block')
            msg = cr.fetchone()
            if msg:
                logger.info('[%s] database suspended: %s', db, msg[0])
                self.abortResponse(None, 'Database suspended', 'warning', msg[0])

            utc = self.UNTIMED_CALLS.get(model, [])
            if utc is not '*' and method not in utc:

                cache = self.access_cache.setdefault(db, {})
                if (uid not in cache) or (cache[uid] < time.time()):
                    try:
                        now = time.strftime('%Y-%m-%d %H:%M:%S')
                        cr.execute("INSERT INTO use_control_time (user_id, date, duration, active) VALUES (%s, %s, %s, %s)", 
                                    (int(uid), now, self.HOUR_MINI, True))
                    except IntegrityError:
                        # if user does not exists do not record the time in the cache
                        pass
                    else:
                        cache[uid] = time.time() + 3600 * self.HOUR_MINI
                    cr.commit()

        finally:
            cr.close()

recording_objects_proxy()

##############################################################################
# Service to request feedback on a database usage
##############################################################################

class use_control_service(netsvc.ExportService):
    def __init__(self, name="use_control"):
        netsvc.ExportService.__init__(self, name)
        self.joinGroup("web-services")

    # WTF is that awful API ?
    def dispatch(self, method, auth, args):
        return getattr(self, method)(*args)

    def new_dispatch(self, method, auth, args):
        pass

    def knock_knock(self):
        return "Who's there?"

    def _get_size(self, cr, dbname):
        """Return the size of the system in Mb."""
        cr.execute('select pg_database_size(%s)', (dbname,))
        db_size = cr.fetchone()[0]
        filestore = os.path.join(tools.config['root_path'], 'filestore', dbname)
        dir_size = sum(os.path.getsize(os.path.join(path, file))
                        for path, _, files in os.walk(filestore)
                        for file in files)
        r = (dir_size + db_size) / (1024.0*1024*1024)   # in GB
        return ceil(r)

    def data_get(self, password, db):
        security.check_super(password)
        cr = pooler.get_db_only(db).cursor()
        try:
            cr.execute('''SELECT to_char(t.date, 'YYYY-MM-DD') as date,
                                 u.name as username,
                                 u.login as login,
                                 sum(t.duration) as hours
                            FROM use_control_time t,
                                 res_users u
                           WHERE u.id = t.user_id
                             AND t.active = %s
                             AND (uploaded = %s OR uploaded IS NULL)
                        GROUP BY to_char(t.date, 'YYYY-MM-DD'),
                                 u.name,
                                 u.login
                       ''', (True, False)
                       )
            data = cr.fetchall()
            cr.execute('UPDATE use_control_time SET uploaded=%s WHERE active = %s AND (uploaded = %s OR uploaded IS NULL)', (True, True, False))
            cr.execute('SELECT name FROM ir_module_module WHERE state=%s', ('installed',))
            modules = map(lambda x: x[0], cr.fetchall())

            hours = reduce(lambda x, y: x+y[3], data, 0.0)

            demo_users = (
                'base.user_demo',
                'project.res_users_analyst',
                'project.res_users_project_manager',
                'project.res_users_technical_leader',
                'project.res_users_developer',
                'project.res_users_designer',
                'project.res_users_tester',
                'project.res_users_finacial_manager',
                'multi_company.res_users_editor_demo',
                'multi_company.res_users_sales_demo_us',
                'multi_company.res_users_sales_demo_be',
                'multi_company.res_users_sales_demo_in',
                'multi_company.res_users_odoo_demo',
            )
            cr.execute('''SELECT count(id)
                            FROM res_users
                           WHERE active = %s
                             AND COALESCE(password, '') != ''
                             AND id not in (SELECT res_id
                                              FROM (SELECT res_id, model, module || '.' || name as xml_id
                                                      FROM ir_model_data) foo
                                             WHERE model = %s
                                               AND xml_id in %s
                                            )
                       ''', (True, 'res.users', demo_users))
            users = cr.fetchone()[0]

            cr.execute('SELECT max(date) FROM use_control_time WHERE active = %s', (True,))
            maxdate = cr.fetchone()[0]
            return {
                'details': data,
                'modules': modules,
                'latest_connection': maxdate or False,
                'users_number': users,
                'space': self._get_size(cr, db),
                'hours': hours,
            }
        finally:
            cr.commit()
            cr.close()

    def block(self, password, dbname, message):
        security.check_super(password)
        cr = pooler.get_db_only(dbname).cursor()
        try:
            cr.execute('INSERT INTO use_control_db_block(name) VALUES(%s)', (message,))
        finally:
            cr.commit()
            cr.close()
        return True

    def unblock(self, password, dbname):
        security.check_super(password)
        cr = pooler.get_db_only(dbname).cursor()
        try:
            cr.execute('DELETE FROM use_control_db_block')
        finally:
            cr.commit()
            cr.close()
        return True

    def clean(self, password, dbname):
        security.check_super(password)
        cr = pooler.get_db_only(dbname).cursor()
        try:
            cr.execute('UPDATE use_control_time SET active = %s', (False,))
        finally:
            cr.commit()
            cr.close()
        return True


    def stats(self, password):
        security.check_super(password)

        try:
            from sql_db import _Pool
            used = len([1 for c, u in _Pool._connections[:] if u])
            count = len(_Pool._connections)
        except ImportError:
            # al pool
            from sql_db import connectionpool
            used = len(connectionpool._used)
            count = used + len(connectionpool._pool)

        active_users = []
        for cache in recording_objects_proxy.access_cache.values():
            au = len([1 for t in cache.values() if t > time.time()])
            active_users.append(au)

        active_dbs = len([1 for u in active_users if u])

        gc.collect()

        return {
            'pooler_length': len(pooler.pool_dic),
            'dbpool_count': count,
            'dbpool_used': used,
            'active_users': active_users,
            'active_dbs': active_dbs,
            'gc_count': len(gc.get_objects()),
            'gc_garbage': len(gc.garbage),
        }

use_control_service()


